syntax = "proto3";

import "google/type/timeofday.proto";

option java_package = "io.confluent.kafka.serializers.protobuf.test";

// Wrapper message for `TimeOfDay`.
message TimeOfDayValue {
  // The bytes value.
  google.type.TimeOfDay value = 1;
}
