syntax = "proto3";

import "confluent/meta.proto";

option java_package = "io.confluent.kafka.serializers.protobuf.test";

// Wrapper message for `Int8`.
message Int8Value {
  int32 value = 1 [(confluent.field_meta) = { params: [
    { key: "connect.type", value: "int8" }
  ]}];
}
