// Generated by the protocol buffer compiler.  DO NOT EDIT!
// source: common_proto.proto

package com.acme.glup;

public final class CommonProto {
  private CommonProto() {}
  public static void registerAllExtensions(
      com.google.protobuf.ExtensionRegistryLite registry) {
  }

  public static void registerAllExtensions(
      com.google.protobuf.ExtensionRegistry registry) {
    registerAllExtensions(
        (com.google.protobuf.ExtensionRegistryLite) registry);
  }
  /**
   * <pre>
   * Describes if an event or campaign is an appinstall or normal
   * retargeting one
   * </pre>
   *
   * Protobuf enum {@code Acme.Glup.MarketingObjectiveType}
   */
  public enum MarketingObjectiveType
      implements com.google.protobuf.ProtocolMessageEnum {
    /**
     * <code>Sale = 0;</code>
     */
    Sale(0),
    /**
     * <code>Install = 1;</code>
     */
    Install(1),
    UNRECOGNIZED(-1),
    ;

    /**
     * <code>Sale = 0;</code>
     */
    public static final int Sale_VALUE = 0;
    /**
     * <code>Install = 1;</code>
     */
    public static final int Install_VALUE = 1;


    public final int getNumber() {
      if (this == UNRECOGNIZED) {
        throw new java.lang.IllegalArgumentException(
            "Can't get the number of an unknown enum value.");
      }
      return value;
    }

    /**
     * @param value The numeric wire value of the corresponding enum entry.
     * @return The enum associated with the given numeric wire value.
     * @deprecated Use {@link #forNumber(int)} instead.
     */
    @java.lang.Deprecated
    public static MarketingObjectiveType valueOf(int value) {
      return forNumber(value);
    }

    /**
     * @param value The numeric wire value of the corresponding enum entry.
     * @return The enum associated with the given numeric wire value.
     */
    public static MarketingObjectiveType forNumber(int value) {
      switch (value) {
        case 0: return Sale;
        case 1: return Install;
        default: return null;
      }
    }

    public static com.google.protobuf.Internal.EnumLiteMap<MarketingObjectiveType>
        internalGetValueMap() {
      return internalValueMap;
    }
    private static final com.google.protobuf.Internal.EnumLiteMap<
        MarketingObjectiveType> internalValueMap =
          new com.google.protobuf.Internal.EnumLiteMap<MarketingObjectiveType>() {
            public MarketingObjectiveType findValueByNumber(int number) {
              return MarketingObjectiveType.forNumber(number);
            }
          };

    public final com.google.protobuf.Descriptors.EnumValueDescriptor
        getValueDescriptor() {
      if (this == UNRECOGNIZED) {
        throw new java.lang.IllegalStateException(
            "Can't get the descriptor of an unrecognized enum value.");
      }
      return getDescriptor().getValues().get(ordinal());
    }
    public final com.google.protobuf.Descriptors.EnumDescriptor
        getDescriptorForType() {
      return getDescriptor();
    }
    public static final com.google.protobuf.Descriptors.EnumDescriptor
        getDescriptor() {
      return com.acme.glup.CommonProto.getDescriptor().getEnumTypes().get(0);
    }

    private static final MarketingObjectiveType[] VALUES = values();

    public static MarketingObjectiveType valueOf(
        com.google.protobuf.Descriptors.EnumValueDescriptor desc) {
      if (desc.getType() != getDescriptor()) {
        throw new java.lang.IllegalArgumentException(
          "EnumValueDescriptor is not for this type.");
      }
      if (desc.getIndex() == -1) {
        return UNRECOGNIZED;
      }
      return VALUES[desc.getIndex()];
    }

    private final int value;

    private MarketingObjectiveType(int value) {
      this.value = value;
    }

    // @@protoc_insertion_point(enum_scope:Acme.Glup.MarketingObjectiveType)
  }

  /**
   * <pre>
   * Enumerate reasons of opt-out
   * sorted by check order
   * </pre>
   *
   * Protobuf enum {@code Acme.Glup.IdentificationForbiddenReason}
   */
  public enum IdentificationForbiddenReason
      implements com.google.protobuf.ProtocolMessageEnum {
    /**
     * <code>NoReason = 0;</code>
     */
    NoReason(0),
    /**
     * <code>ExplicitConsentRequired = 1;</code>
     */
    ExplicitConsentRequired(1),
    /**
     * <code>OptoutCookie = 2;</code>
     */
    OptoutCookie(2),
    /**
     * <code>CtoOptoutCookie = 3;</code>
     */
    CtoOptoutCookie(3),
    /**
     * <code>LimitedAdTracking = 4;</code>
     */
    LimitedAdTracking(4),
    /**
     * <code>HstsOptout = 5;</code>
     */
    HstsOptout(5),
    /**
     * <code>DoNotTrackHeader = 6;</code>
     */
    DoNotTrackHeader(6),
    /**
     * <code>OoOCookie = 7;</code>
     */
    OoOCookie(7),
    /**
     * <code>PendingOptout = 8;</code>
     */
    PendingOptout(8),
    UNRECOGNIZED(-1),
    ;

    /**
     * <code>NoReason = 0;</code>
     */
    public static final int NoReason_VALUE = 0;
    /**
     * <code>ExplicitConsentRequired = 1;</code>
     */
    public static final int ExplicitConsentRequired_VALUE = 1;
    /**
     * <code>OptoutCookie = 2;</code>
     */
    public static final int OptoutCookie_VALUE = 2;
    /**
     * <code>CtoOptoutCookie = 3;</code>
     */
    public static final int CtoOptoutCookie_VALUE = 3;
    /**
     * <code>LimitedAdTracking = 4;</code>
     */
    public static final int LimitedAdTracking_VALUE = 4;
    /**
     * <code>HstsOptout = 5;</code>
     */
    public static final int HstsOptout_VALUE = 5;
    /**
     * <code>DoNotTrackHeader = 6;</code>
     */
    public static final int DoNotTrackHeader_VALUE = 6;
    /**
     * <code>OoOCookie = 7;</code>
     */
    public static final int OoOCookie_VALUE = 7;
    /**
     * <code>PendingOptout = 8;</code>
     */
    public static final int PendingOptout_VALUE = 8;


    public final int getNumber() {
      if (this == UNRECOGNIZED) {
        throw new java.lang.IllegalArgumentException(
            "Can't get the number of an unknown enum value.");
      }
      return value;
    }

    /**
     * @param value The numeric wire value of the corresponding enum entry.
     * @return The enum associated with the given numeric wire value.
     * @deprecated Use {@link #forNumber(int)} instead.
     */
    @java.lang.Deprecated
    public static IdentificationForbiddenReason valueOf(int value) {
      return forNumber(value);
    }

    /**
     * @param value The numeric wire value of the corresponding enum entry.
     * @return The enum associated with the given numeric wire value.
     */
    public static IdentificationForbiddenReason forNumber(int value) {
      switch (value) {
        case 0: return NoReason;
        case 1: return ExplicitConsentRequired;
        case 2: return OptoutCookie;
        case 3: return CtoOptoutCookie;
        case 4: return LimitedAdTracking;
        case 5: return HstsOptout;
        case 6: return DoNotTrackHeader;
        case 7: return OoOCookie;
        case 8: return PendingOptout;
        default: return null;
      }
    }

    public static com.google.protobuf.Internal.EnumLiteMap<IdentificationForbiddenReason>
        internalGetValueMap() {
      return internalValueMap;
    }
    private static final com.google.protobuf.Internal.EnumLiteMap<
        IdentificationForbiddenReason> internalValueMap =
          new com.google.protobuf.Internal.EnumLiteMap<IdentificationForbiddenReason>() {
            public IdentificationForbiddenReason findValueByNumber(int number) {
              return IdentificationForbiddenReason.forNumber(number);
            }
          };

    public final com.google.protobuf.Descriptors.EnumValueDescriptor
        getValueDescriptor() {
      if (this == UNRECOGNIZED) {
        throw new java.lang.IllegalStateException(
            "Can't get the descriptor of an unrecognized enum value.");
      }
      return getDescriptor().getValues().get(ordinal());
    }
    public final com.google.protobuf.Descriptors.EnumDescriptor
        getDescriptorForType() {
      return getDescriptor();
    }
    public static final com.google.protobuf.Descriptors.EnumDescriptor
        getDescriptor() {
      return com.acme.glup.CommonProto.getDescriptor().getEnumTypes().get(1);
    }

    private static final IdentificationForbiddenReason[] VALUES = values();

    public static IdentificationForbiddenReason valueOf(
        com.google.protobuf.Descriptors.EnumValueDescriptor desc) {
      if (desc.getType() != getDescriptor()) {
        throw new java.lang.IllegalArgumentException(
          "EnumValueDescriptor is not for this type.");
      }
      if (desc.getIndex() == -1) {
        return UNRECOGNIZED;
      }
      return VALUES[desc.getIndex()];
    }

    private final int value;

    private IdentificationForbiddenReason(int value) {
      this.value = value;
    }

    // @@protoc_insertion_point(enum_scope:Acme.Glup.IdentificationForbiddenReason)
  }

  public interface ConsentOrBuilder extends
      // @@protoc_insertion_point(interface_extends:Acme.Glup.Consent)
      com.google.protobuf.MessageOrBuilder {

    /**
     * <pre>
     *Value of IdentificationForbidden for user identity
     * </pre>
     *
     * <code>bool identification_forbidden = 1;</code>
     * @return The identificationForbidden.
     */
    boolean getIdentificationForbidden();

    /**
     * <pre>
     *Reason for identification forbidden, NoReason if identification not forbidden
     * </pre>
     *
     * <code>.Acme.Glup.IdentificationForbiddenReason reason = 2;</code>
     * @return The enum numeric value on the wire for reason.
     */
    int getReasonValue();
    /**
     * <pre>
     *Reason for identification forbidden, NoReason if identification not forbidden
     * </pre>
     *
     * <code>.Acme.Glup.IdentificationForbiddenReason reason = 2;</code>
     * @return The reason.
     */
    com.acme.glup.CommonProto.IdentificationForbiddenReason getReason();

    /**
     * <pre>
     * nullable tracking special field
     * </pre>
     *
     * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
     */
    int getSetFieldsCount();
    /**
     * <pre>
     * nullable tracking special field
     * </pre>
     *
     * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
     */
    boolean containsSetFields(
        int key);
    /**
     * Use {@link #getSetFieldsMap()} instead.
     */
    @java.lang.Deprecated
    java.util.Map<java.lang.Integer, java.lang.Boolean>
    getSetFields();
    /**
     * <pre>
     * nullable tracking special field
     * </pre>
     *
     * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
     */
    java.util.Map<java.lang.Integer, java.lang.Boolean>
    getSetFieldsMap();
    /**
     * <pre>
     * nullable tracking special field
     * </pre>
     *
     * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
     */

    boolean getSetFieldsOrDefault(
        int key,
        boolean defaultValue);
    /**
     * <pre>
     * nullable tracking special field
     * </pre>
     *
     * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
     */

    boolean getSetFieldsOrThrow(
        int key);
  }
  /**
   * <pre>
   * Describes the consent status and the reasons
   * of consent refusal
   * </pre>
   *
   * Protobuf type {@code Acme.Glup.Consent}
   */
  public static final class Consent extends
      com.google.protobuf.GeneratedMessageV3 implements
      // @@protoc_insertion_point(message_implements:Acme.Glup.Consent)
      ConsentOrBuilder {
  private static final long serialVersionUID = 0L;
    // Use Consent.newBuilder() to construct.
    private Consent(com.google.protobuf.GeneratedMessageV3.Builder<?> builder) {
      super(builder);
    }
    private Consent() {
      reason_ = 0;
    }

    @java.lang.Override
    @SuppressWarnings({"unused"})
    protected java.lang.Object newInstance(
        UnusedPrivateParameter unused) {
      return new Consent();
    }

    @java.lang.Override
    public final com.google.protobuf.UnknownFieldSet
    getUnknownFields() {
      return this.unknownFields;
    }
    private Consent(
        com.google.protobuf.CodedInputStream input,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws com.google.protobuf.InvalidProtocolBufferException {
      this();
      if (extensionRegistry == null) {
        throw new java.lang.NullPointerException();
      }
      int mutable_bitField0_ = 0;
      com.google.protobuf.UnknownFieldSet.Builder unknownFields =
          com.google.protobuf.UnknownFieldSet.newBuilder();
      try {
        boolean done = false;
        while (!done) {
          int tag = input.readTag();
          switch (tag) {
            case 0:
              done = true;
              break;
            case 8: {

              identificationForbidden_ = input.readBool();
              break;
            }
            case 16: {
              int rawValue = input.readEnum();

              reason_ = rawValue;
              break;
            }
            case 400082: {
              if (!((mutable_bitField0_ & 0x00000001) != 0)) {
                setFields_ = com.google.protobuf.MapField.newMapField(
                    SetFieldsDefaultEntryHolder.defaultEntry);
                mutable_bitField0_ |= 0x00000001;
              }
              com.google.protobuf.MapEntry<java.lang.Integer, java.lang.Boolean>
              setFields__ = input.readMessage(
                  SetFieldsDefaultEntryHolder.defaultEntry.getParserForType(), extensionRegistry);
              setFields_.getMutableMap().put(
                  setFields__.getKey(), setFields__.getValue());
              break;
            }
            default: {
              if (!parseUnknownField(
                  input, unknownFields, extensionRegistry, tag)) {
                done = true;
              }
              break;
            }
          }
        }
      } catch (com.google.protobuf.InvalidProtocolBufferException e) {
        throw e.setUnfinishedMessage(this);
      } catch (java.io.IOException e) {
        throw new com.google.protobuf.InvalidProtocolBufferException(
            e).setUnfinishedMessage(this);
      } finally {
        this.unknownFields = unknownFields.build();
        makeExtensionsImmutable();
      }
    }
    public static final com.google.protobuf.Descriptors.Descriptor
        getDescriptor() {
      return com.acme.glup.CommonProto.internal_static_Acme_Glup_Consent_descriptor;
    }

    @SuppressWarnings({"rawtypes"})
    @java.lang.Override
    protected com.google.protobuf.MapField internalGetMapField(
        int number) {
      switch (number) {
        case 50010:
          return internalGetSetFields();
        default:
          throw new RuntimeException(
              "Invalid map field number: " + number);
      }
    }
    @java.lang.Override
    protected com.google.protobuf.GeneratedMessageV3.FieldAccessorTable
        internalGetFieldAccessorTable() {
      return com.acme.glup.CommonProto.internal_static_Acme_Glup_Consent_fieldAccessorTable
          .ensureFieldAccessorsInitialized(
              com.acme.glup.CommonProto.Consent.class, com.acme.glup.CommonProto.Consent.Builder.class);
    }

    public static final int IDENTIFICATION_FORBIDDEN_FIELD_NUMBER = 1;
    private boolean identificationForbidden_;
    /**
     * <pre>
     *Value of IdentificationForbidden for user identity
     * </pre>
     *
     * <code>bool identification_forbidden = 1;</code>
     * @return The identificationForbidden.
     */
    @java.lang.Override
    public boolean getIdentificationForbidden() {
      return identificationForbidden_;
    }

    public static final int REASON_FIELD_NUMBER = 2;
    private int reason_;
    /**
     * <pre>
     *Reason for identification forbidden, NoReason if identification not forbidden
     * </pre>
     *
     * <code>.Acme.Glup.IdentificationForbiddenReason reason = 2;</code>
     * @return The enum numeric value on the wire for reason.
     */
    @java.lang.Override public int getReasonValue() {
      return reason_;
    }
    /**
     * <pre>
     *Reason for identification forbidden, NoReason if identification not forbidden
     * </pre>
     *
     * <code>.Acme.Glup.IdentificationForbiddenReason reason = 2;</code>
     * @return The reason.
     */
    @java.lang.Override public com.acme.glup.CommonProto.IdentificationForbiddenReason getReason() {
      @SuppressWarnings("deprecation")
      com.acme.glup.CommonProto.IdentificationForbiddenReason result = com.acme.glup.CommonProto.IdentificationForbiddenReason.valueOf(reason_);
      return result == null ? com.acme.glup.CommonProto.IdentificationForbiddenReason.UNRECOGNIZED : result;
    }

    public static final int SET_FIELDS_FIELD_NUMBER = 50010;
    private static final class SetFieldsDefaultEntryHolder {
      static final com.google.protobuf.MapEntry<
          java.lang.Integer, java.lang.Boolean> defaultEntry =
              com.google.protobuf.MapEntry
              .<java.lang.Integer, java.lang.Boolean>newDefaultInstance(
                  com.acme.glup.CommonProto.internal_static_Acme_Glup_Consent_SetFieldsEntry_descriptor, 
                  com.google.protobuf.WireFormat.FieldType.INT32,
                  0,
                  com.google.protobuf.WireFormat.FieldType.BOOL,
                  false);
    }
    private com.google.protobuf.MapField<
        java.lang.Integer, java.lang.Boolean> setFields_;
    private com.google.protobuf.MapField<java.lang.Integer, java.lang.Boolean>
    internalGetSetFields() {
      if (setFields_ == null) {
        return com.google.protobuf.MapField.emptyMapField(
            SetFieldsDefaultEntryHolder.defaultEntry);
      }
      return setFields_;
    }

    public int getSetFieldsCount() {
      return internalGetSetFields().getMap().size();
    }
    /**
     * <pre>
     * nullable tracking special field
     * </pre>
     *
     * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
     */

    @java.lang.Override
    public boolean containsSetFields(
        int key) {
      
      return internalGetSetFields().getMap().containsKey(key);
    }
    /**
     * Use {@link #getSetFieldsMap()} instead.
     */
    @java.lang.Override
    @java.lang.Deprecated
    public java.util.Map<java.lang.Integer, java.lang.Boolean> getSetFields() {
      return getSetFieldsMap();
    }
    /**
     * <pre>
     * nullable tracking special field
     * </pre>
     *
     * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
     */
    @java.lang.Override

    public java.util.Map<java.lang.Integer, java.lang.Boolean> getSetFieldsMap() {
      return internalGetSetFields().getMap();
    }
    /**
     * <pre>
     * nullable tracking special field
     * </pre>
     *
     * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
     */
    @java.lang.Override

    public boolean getSetFieldsOrDefault(
        int key,
        boolean defaultValue) {
      
      java.util.Map<java.lang.Integer, java.lang.Boolean> map =
          internalGetSetFields().getMap();
      return map.containsKey(key) ? map.get(key) : defaultValue;
    }
    /**
     * <pre>
     * nullable tracking special field
     * </pre>
     *
     * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
     */
    @java.lang.Override

    public boolean getSetFieldsOrThrow(
        int key) {
      
      java.util.Map<java.lang.Integer, java.lang.Boolean> map =
          internalGetSetFields().getMap();
      if (!map.containsKey(key)) {
        throw new java.lang.IllegalArgumentException();
      }
      return map.get(key);
    }

    private byte memoizedIsInitialized = -1;
    @java.lang.Override
    public final boolean isInitialized() {
      byte isInitialized = memoizedIsInitialized;
      if (isInitialized == 1) return true;
      if (isInitialized == 0) return false;

      memoizedIsInitialized = 1;
      return true;
    }

    @java.lang.Override
    public void writeTo(com.google.protobuf.CodedOutputStream output)
                        throws java.io.IOException {
      if (identificationForbidden_ != false) {
        output.writeBool(1, identificationForbidden_);
      }
      if (reason_ != com.acme.glup.CommonProto.IdentificationForbiddenReason.NoReason.getNumber()) {
        output.writeEnum(2, reason_);
      }
      com.google.protobuf.GeneratedMessageV3
        .serializeIntegerMapTo(
          output,
          internalGetSetFields(),
          SetFieldsDefaultEntryHolder.defaultEntry,
          50010);
      unknownFields.writeTo(output);
    }

    @java.lang.Override
    public int getSerializedSize() {
      int size = memoizedSize;
      if (size != -1) return size;

      size = 0;
      if (identificationForbidden_ != false) {
        size += com.google.protobuf.CodedOutputStream
          .computeBoolSize(1, identificationForbidden_);
      }
      if (reason_ != com.acme.glup.CommonProto.IdentificationForbiddenReason.NoReason.getNumber()) {
        size += com.google.protobuf.CodedOutputStream
          .computeEnumSize(2, reason_);
      }
      for (java.util.Map.Entry<java.lang.Integer, java.lang.Boolean> entry
           : internalGetSetFields().getMap().entrySet()) {
        com.google.protobuf.MapEntry<java.lang.Integer, java.lang.Boolean>
        setFields__ = SetFieldsDefaultEntryHolder.defaultEntry.newBuilderForType()
            .setKey(entry.getKey())
            .setValue(entry.getValue())
            .build();
        size += com.google.protobuf.CodedOutputStream
            .computeMessageSize(50010, setFields__);
      }
      size += unknownFields.getSerializedSize();
      memoizedSize = size;
      return size;
    }

    @java.lang.Override
    public boolean equals(final java.lang.Object obj) {
      if (obj == this) {
       return true;
      }
      if (!(obj instanceof com.acme.glup.CommonProto.Consent)) {
        return super.equals(obj);
      }
      com.acme.glup.CommonProto.Consent other = (com.acme.glup.CommonProto.Consent) obj;

      if (getIdentificationForbidden()
          != other.getIdentificationForbidden()) return false;
      if (reason_ != other.reason_) return false;
      if (!internalGetSetFields().equals(
          other.internalGetSetFields())) return false;
      if (!unknownFields.equals(other.unknownFields)) return false;
      return true;
    }

    @java.lang.Override
    public int hashCode() {
      if (memoizedHashCode != 0) {
        return memoizedHashCode;
      }
      int hash = 41;
      hash = (19 * hash) + getDescriptor().hashCode();
      hash = (37 * hash) + IDENTIFICATION_FORBIDDEN_FIELD_NUMBER;
      hash = (53 * hash) + com.google.protobuf.Internal.hashBoolean(
          getIdentificationForbidden());
      hash = (37 * hash) + REASON_FIELD_NUMBER;
      hash = (53 * hash) + reason_;
      if (!internalGetSetFields().getMap().isEmpty()) {
        hash = (37 * hash) + SET_FIELDS_FIELD_NUMBER;
        hash = (53 * hash) + internalGetSetFields().hashCode();
      }
      hash = (29 * hash) + unknownFields.hashCode();
      memoizedHashCode = hash;
      return hash;
    }

    public static com.acme.glup.CommonProto.Consent parseFrom(
        java.nio.ByteBuffer data)
        throws com.google.protobuf.InvalidProtocolBufferException {
      return PARSER.parseFrom(data);
    }
    public static com.acme.glup.CommonProto.Consent parseFrom(
        java.nio.ByteBuffer data,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws com.google.protobuf.InvalidProtocolBufferException {
      return PARSER.parseFrom(data, extensionRegistry);
    }
    public static com.acme.glup.CommonProto.Consent parseFrom(
        com.google.protobuf.ByteString data)
        throws com.google.protobuf.InvalidProtocolBufferException {
      return PARSER.parseFrom(data);
    }
    public static com.acme.glup.CommonProto.Consent parseFrom(
        com.google.protobuf.ByteString data,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws com.google.protobuf.InvalidProtocolBufferException {
      return PARSER.parseFrom(data, extensionRegistry);
    }
    public static com.acme.glup.CommonProto.Consent parseFrom(byte[] data)
        throws com.google.protobuf.InvalidProtocolBufferException {
      return PARSER.parseFrom(data);
    }
    public static com.acme.glup.CommonProto.Consent parseFrom(
        byte[] data,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws com.google.protobuf.InvalidProtocolBufferException {
      return PARSER.parseFrom(data, extensionRegistry);
    }
    public static com.acme.glup.CommonProto.Consent parseFrom(java.io.InputStream input)
        throws java.io.IOException {
      return com.google.protobuf.GeneratedMessageV3
          .parseWithIOException(PARSER, input);
    }
    public static com.acme.glup.CommonProto.Consent parseFrom(
        java.io.InputStream input,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws java.io.IOException {
      return com.google.protobuf.GeneratedMessageV3
          .parseWithIOException(PARSER, input, extensionRegistry);
    }
    public static com.acme.glup.CommonProto.Consent parseDelimitedFrom(java.io.InputStream input)
        throws java.io.IOException {
      return com.google.protobuf.GeneratedMessageV3
          .parseDelimitedWithIOException(PARSER, input);
    }
    public static com.acme.glup.CommonProto.Consent parseDelimitedFrom(
        java.io.InputStream input,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws java.io.IOException {
      return com.google.protobuf.GeneratedMessageV3
          .parseDelimitedWithIOException(PARSER, input, extensionRegistry);
    }
    public static com.acme.glup.CommonProto.Consent parseFrom(
        com.google.protobuf.CodedInputStream input)
        throws java.io.IOException {
      return com.google.protobuf.GeneratedMessageV3
          .parseWithIOException(PARSER, input);
    }
    public static com.acme.glup.CommonProto.Consent parseFrom(
        com.google.protobuf.CodedInputStream input,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws java.io.IOException {
      return com.google.protobuf.GeneratedMessageV3
          .parseWithIOException(PARSER, input, extensionRegistry);
    }

    @java.lang.Override
    public Builder newBuilderForType() { return newBuilder(); }
    public static Builder newBuilder() {
      return DEFAULT_INSTANCE.toBuilder();
    }
    public static Builder newBuilder(com.acme.glup.CommonProto.Consent prototype) {
      return DEFAULT_INSTANCE.toBuilder().mergeFrom(prototype);
    }
    @java.lang.Override
    public Builder toBuilder() {
      return this == DEFAULT_INSTANCE
          ? new Builder() : new Builder().mergeFrom(this);
    }

    @java.lang.Override
    protected Builder newBuilderForType(
        com.google.protobuf.GeneratedMessageV3.BuilderParent parent) {
      Builder builder = new Builder(parent);
      return builder;
    }
    /**
     * <pre>
     * Describes the consent status and the reasons
     * of consent refusal
     * </pre>
     *
     * Protobuf type {@code Acme.Glup.Consent}
     */
    public static final class Builder extends
        com.google.protobuf.GeneratedMessageV3.Builder<Builder> implements
        // @@protoc_insertion_point(builder_implements:Acme.Glup.Consent)
        com.acme.glup.CommonProto.ConsentOrBuilder {
      public static final com.google.protobuf.Descriptors.Descriptor
          getDescriptor() {
        return com.acme.glup.CommonProto.internal_static_Acme_Glup_Consent_descriptor;
      }

      @SuppressWarnings({"rawtypes"})
      protected com.google.protobuf.MapField internalGetMapField(
          int number) {
        switch (number) {
          case 50010:
            return internalGetSetFields();
          default:
            throw new RuntimeException(
                "Invalid map field number: " + number);
        }
      }
      @SuppressWarnings({"rawtypes"})
      protected com.google.protobuf.MapField internalGetMutableMapField(
          int number) {
        switch (number) {
          case 50010:
            return internalGetMutableSetFields();
          default:
            throw new RuntimeException(
                "Invalid map field number: " + number);
        }
      }
      @java.lang.Override
      protected com.google.protobuf.GeneratedMessageV3.FieldAccessorTable
          internalGetFieldAccessorTable() {
        return com.acme.glup.CommonProto.internal_static_Acme_Glup_Consent_fieldAccessorTable
            .ensureFieldAccessorsInitialized(
                com.acme.glup.CommonProto.Consent.class, com.acme.glup.CommonProto.Consent.Builder.class);
      }

      // Construct using com.acme.glup.CommonProto.Consent.newBuilder()
      private Builder() {
        maybeForceBuilderInitialization();
      }

      private Builder(
          com.google.protobuf.GeneratedMessageV3.BuilderParent parent) {
        super(parent);
        maybeForceBuilderInitialization();
      }
      private void maybeForceBuilderInitialization() {
        if (com.google.protobuf.GeneratedMessageV3
                .alwaysUseFieldBuilders) {
        }
      }
      @java.lang.Override
      public Builder clear() {
        super.clear();
        identificationForbidden_ = false;

        reason_ = 0;

        internalGetMutableSetFields().clear();
        return this;
      }

      @java.lang.Override
      public com.google.protobuf.Descriptors.Descriptor
          getDescriptorForType() {
        return com.acme.glup.CommonProto.internal_static_Acme_Glup_Consent_descriptor;
      }

      @java.lang.Override
      public com.acme.glup.CommonProto.Consent getDefaultInstanceForType() {
        return com.acme.glup.CommonProto.Consent.getDefaultInstance();
      }

      @java.lang.Override
      public com.acme.glup.CommonProto.Consent build() {
        com.acme.glup.CommonProto.Consent result = buildPartial();
        if (!result.isInitialized()) {
          throw newUninitializedMessageException(result);
        }
        return result;
      }

      @java.lang.Override
      public com.acme.glup.CommonProto.Consent buildPartial() {
        com.acme.glup.CommonProto.Consent result = new com.acme.glup.CommonProto.Consent(this);
        int from_bitField0_ = bitField0_;
        result.identificationForbidden_ = identificationForbidden_;
        result.reason_ = reason_;
        result.setFields_ = internalGetSetFields();
        result.setFields_.makeImmutable();
        onBuilt();
        return result;
      }

      @java.lang.Override
      public Builder clone() {
        return super.clone();
      }
      @java.lang.Override
      public Builder setField(
          com.google.protobuf.Descriptors.FieldDescriptor field,
          java.lang.Object value) {
        return super.setField(field, value);
      }
      @java.lang.Override
      public Builder clearField(
          com.google.protobuf.Descriptors.FieldDescriptor field) {
        return super.clearField(field);
      }
      @java.lang.Override
      public Builder clearOneof(
          com.google.protobuf.Descriptors.OneofDescriptor oneof) {
        return super.clearOneof(oneof);
      }
      @java.lang.Override
      public Builder setRepeatedField(
          com.google.protobuf.Descriptors.FieldDescriptor field,
          int index, java.lang.Object value) {
        return super.setRepeatedField(field, index, value);
      }
      @java.lang.Override
      public Builder addRepeatedField(
          com.google.protobuf.Descriptors.FieldDescriptor field,
          java.lang.Object value) {
        return super.addRepeatedField(field, value);
      }
      @java.lang.Override
      public Builder mergeFrom(com.google.protobuf.Message other) {
        if (other instanceof com.acme.glup.CommonProto.Consent) {
          return mergeFrom((com.acme.glup.CommonProto.Consent)other);
        } else {
          super.mergeFrom(other);
          return this;
        }
      }

      public Builder mergeFrom(com.acme.glup.CommonProto.Consent other) {
        if (other == com.acme.glup.CommonProto.Consent.getDefaultInstance()) return this;
        if (other.getIdentificationForbidden() != false) {
          setIdentificationForbidden(other.getIdentificationForbidden());
        }
        if (other.reason_ != 0) {
          setReasonValue(other.getReasonValue());
        }
        internalGetMutableSetFields().mergeFrom(
            other.internalGetSetFields());
        this.mergeUnknownFields(other.unknownFields);
        onChanged();
        return this;
      }

      @java.lang.Override
      public final boolean isInitialized() {
        return true;
      }

      @java.lang.Override
      public Builder mergeFrom(
          com.google.protobuf.CodedInputStream input,
          com.google.protobuf.ExtensionRegistryLite extensionRegistry)
          throws java.io.IOException {
        com.acme.glup.CommonProto.Consent parsedMessage = null;
        try {
          parsedMessage = PARSER.parsePartialFrom(input, extensionRegistry);
        } catch (com.google.protobuf.InvalidProtocolBufferException e) {
          parsedMessage = (com.acme.glup.CommonProto.Consent) e.getUnfinishedMessage();
          throw e.unwrapIOException();
        } finally {
          if (parsedMessage != null) {
            mergeFrom(parsedMessage);
          }
        }
        return this;
      }
      private int bitField0_;

      private boolean identificationForbidden_ ;
      /**
       * <pre>
       *Value of IdentificationForbidden for user identity
       * </pre>
       *
       * <code>bool identification_forbidden = 1;</code>
       * @return The identificationForbidden.
       */
      @java.lang.Override
      public boolean getIdentificationForbidden() {
        return identificationForbidden_;
      }
      /**
       * <pre>
       *Value of IdentificationForbidden for user identity
       * </pre>
       *
       * <code>bool identification_forbidden = 1;</code>
       * @param value The identificationForbidden to set.
       * @return This builder for chaining.
       */
      public Builder setIdentificationForbidden(boolean value) {
        
        identificationForbidden_ = value;
        onChanged();
        return this;
      }
      /**
       * <pre>
       *Value of IdentificationForbidden for user identity
       * </pre>
       *
       * <code>bool identification_forbidden = 1;</code>
       * @return This builder for chaining.
       */
      public Builder clearIdentificationForbidden() {
        
        identificationForbidden_ = false;
        onChanged();
        return this;
      }

      private int reason_ = 0;
      /**
       * <pre>
       *Reason for identification forbidden, NoReason if identification not forbidden
       * </pre>
       *
       * <code>.Acme.Glup.IdentificationForbiddenReason reason = 2;</code>
       * @return The enum numeric value on the wire for reason.
       */
      @java.lang.Override public int getReasonValue() {
        return reason_;
      }
      /**
       * <pre>
       *Reason for identification forbidden, NoReason if identification not forbidden
       * </pre>
       *
       * <code>.Acme.Glup.IdentificationForbiddenReason reason = 2;</code>
       * @param value The enum numeric value on the wire for reason to set.
       * @return This builder for chaining.
       */
      public Builder setReasonValue(int value) {
        
        reason_ = value;
        onChanged();
        return this;
      }
      /**
       * <pre>
       *Reason for identification forbidden, NoReason if identification not forbidden
       * </pre>
       *
       * <code>.Acme.Glup.IdentificationForbiddenReason reason = 2;</code>
       * @return The reason.
       */
      @java.lang.Override
      public com.acme.glup.CommonProto.IdentificationForbiddenReason getReason() {
        @SuppressWarnings("deprecation")
        com.acme.glup.CommonProto.IdentificationForbiddenReason result = com.acme.glup.CommonProto.IdentificationForbiddenReason.valueOf(reason_);
        return result == null ? com.acme.glup.CommonProto.IdentificationForbiddenReason.UNRECOGNIZED : result;
      }
      /**
       * <pre>
       *Reason for identification forbidden, NoReason if identification not forbidden
       * </pre>
       *
       * <code>.Acme.Glup.IdentificationForbiddenReason reason = 2;</code>
       * @param value The reason to set.
       * @return This builder for chaining.
       */
      public Builder setReason(com.acme.glup.CommonProto.IdentificationForbiddenReason value) {
        if (value == null) {
          throw new NullPointerException();
        }
        
        reason_ = value.getNumber();
        onChanged();
        return this;
      }
      /**
       * <pre>
       *Reason for identification forbidden, NoReason if identification not forbidden
       * </pre>
       *
       * <code>.Acme.Glup.IdentificationForbiddenReason reason = 2;</code>
       * @return This builder for chaining.
       */
      public Builder clearReason() {
        
        reason_ = 0;
        onChanged();
        return this;
      }

      private com.google.protobuf.MapField<
          java.lang.Integer, java.lang.Boolean> setFields_;
      private com.google.protobuf.MapField<java.lang.Integer, java.lang.Boolean>
      internalGetSetFields() {
        if (setFields_ == null) {
          return com.google.protobuf.MapField.emptyMapField(
              SetFieldsDefaultEntryHolder.defaultEntry);
        }
        return setFields_;
      }
      private com.google.protobuf.MapField<java.lang.Integer, java.lang.Boolean>
      internalGetMutableSetFields() {
        onChanged();;
        if (setFields_ == null) {
          setFields_ = com.google.protobuf.MapField.newMapField(
              SetFieldsDefaultEntryHolder.defaultEntry);
        }
        if (!setFields_.isMutable()) {
          setFields_ = setFields_.copy();
        }
        return setFields_;
      }

      public int getSetFieldsCount() {
        return internalGetSetFields().getMap().size();
      }
      /**
       * <pre>
       * nullable tracking special field
       * </pre>
       *
       * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
       */

      @java.lang.Override
      public boolean containsSetFields(
          int key) {
        
        return internalGetSetFields().getMap().containsKey(key);
      }
      /**
       * Use {@link #getSetFieldsMap()} instead.
       */
      @java.lang.Override
      @java.lang.Deprecated
      public java.util.Map<java.lang.Integer, java.lang.Boolean> getSetFields() {
        return getSetFieldsMap();
      }
      /**
       * <pre>
       * nullable tracking special field
       * </pre>
       *
       * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
       */
      @java.lang.Override

      public java.util.Map<java.lang.Integer, java.lang.Boolean> getSetFieldsMap() {
        return internalGetSetFields().getMap();
      }
      /**
       * <pre>
       * nullable tracking special field
       * </pre>
       *
       * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
       */
      @java.lang.Override

      public boolean getSetFieldsOrDefault(
          int key,
          boolean defaultValue) {
        
        java.util.Map<java.lang.Integer, java.lang.Boolean> map =
            internalGetSetFields().getMap();
        return map.containsKey(key) ? map.get(key) : defaultValue;
      }
      /**
       * <pre>
       * nullable tracking special field
       * </pre>
       *
       * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
       */
      @java.lang.Override

      public boolean getSetFieldsOrThrow(
          int key) {
        
        java.util.Map<java.lang.Integer, java.lang.Boolean> map =
            internalGetSetFields().getMap();
        if (!map.containsKey(key)) {
          throw new java.lang.IllegalArgumentException();
        }
        return map.get(key);
      }

      public Builder clearSetFields() {
        internalGetMutableSetFields().getMutableMap()
            .clear();
        return this;
      }
      /**
       * <pre>
       * nullable tracking special field
       * </pre>
       *
       * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
       */

      public Builder removeSetFields(
          int key) {
        
        internalGetMutableSetFields().getMutableMap()
            .remove(key);
        return this;
      }
      /**
       * Use alternate mutation accessors instead.
       */
      @java.lang.Deprecated
      public java.util.Map<java.lang.Integer, java.lang.Boolean>
      getMutableSetFields() {
        return internalGetMutableSetFields().getMutableMap();
      }
      /**
       * <pre>
       * nullable tracking special field
       * </pre>
       *
       * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
       */
      public Builder putSetFields(
          int key,
          boolean value) {
        
        
        internalGetMutableSetFields().getMutableMap()
            .put(key, value);
        return this;
      }
      /**
       * <pre>
       * nullable tracking special field
       * </pre>
       *
       * <code>map&lt;int32, bool&gt; set_fields = 50010;</code>
       */

      public Builder putAllSetFields(
          java.util.Map<java.lang.Integer, java.lang.Boolean> values) {
        internalGetMutableSetFields().getMutableMap()
            .putAll(values);
        return this;
      }
      @java.lang.Override
      public final Builder setUnknownFields(
          final com.google.protobuf.UnknownFieldSet unknownFields) {
        return super.setUnknownFields(unknownFields);
      }

      @java.lang.Override
      public final Builder mergeUnknownFields(
          final com.google.protobuf.UnknownFieldSet unknownFields) {
        return super.mergeUnknownFields(unknownFields);
      }


      // @@protoc_insertion_point(builder_scope:Acme.Glup.Consent)
    }

    // @@protoc_insertion_point(class_scope:Acme.Glup.Consent)
    private static final com.acme.glup.CommonProto.Consent DEFAULT_INSTANCE;
    static {
      DEFAULT_INSTANCE = new com.acme.glup.CommonProto.Consent();
    }

    public static com.acme.glup.CommonProto.Consent getDefaultInstance() {
      return DEFAULT_INSTANCE;
    }

    private static final com.google.protobuf.Parser<Consent>
        PARSER = new com.google.protobuf.AbstractParser<Consent>() {
      @java.lang.Override
      public Consent parsePartialFrom(
          com.google.protobuf.CodedInputStream input,
          com.google.protobuf.ExtensionRegistryLite extensionRegistry)
          throws com.google.protobuf.InvalidProtocolBufferException {
        return new Consent(input, extensionRegistry);
      }
    };

    public static com.google.protobuf.Parser<Consent> parser() {
      return PARSER;
    }

    @java.lang.Override
    public com.google.protobuf.Parser<Consent> getParserForType() {
      return PARSER;
    }

    @java.lang.Override
    public com.acme.glup.CommonProto.Consent getDefaultInstanceForType() {
      return DEFAULT_INSTANCE;
    }

  }

  private static final com.google.protobuf.Descriptors.Descriptor
    internal_static_Acme_Glup_Consent_descriptor;
  private static final 
    com.google.protobuf.GeneratedMessageV3.FieldAccessorTable
      internal_static_Acme_Glup_Consent_fieldAccessorTable;
  private static final com.google.protobuf.Descriptors.Descriptor
    internal_static_Acme_Glup_Consent_SetFieldsEntry_descriptor;
  private static final 
    com.google.protobuf.GeneratedMessageV3.FieldAccessorTable
      internal_static_Acme_Glup_Consent_SetFieldsEntry_fieldAccessorTable;

  public static com.google.protobuf.Descriptors.FileDescriptor
      getDescriptor() {
    return descriptor;
  }
  private static  com.google.protobuf.Descriptors.FileDescriptor
      descriptor;
  static {
    java.lang.String[] descriptorData = {
      "\n\022common_proto.proto\022\tAcme.Glup\032\024metadat" +
      "a_proto.proto\"\326\001\n\007Consent\022 \n\030identificat" +
      "ion_forbidden\030\001 \001(\010\0228\n\006reason\030\002 \001(\0162(.Ac" +
      "me.Glup.IdentificationForbiddenReason\0227\n" +
      "\nset_fields\030\332\206\003 \003(\0132!.Acme.Glup.Consent." +
      "SetFieldsEntry\0320\n\016SetFieldsEntry\022\013\n\003key\030" +
      "\001 \001(\005\022\r\n\005value\030\002 \001(\010:\0028\001:\004\210\265\030\001*/\n\026Market" +
      "ingObjectiveType\022\010\n\004Sale\020\000\022\013\n\007Install\020\001*" +
      "\320\001\n\035IdentificationForbiddenReason\022\014\n\010NoR" +
      "eason\020\000\022\033\n\027ExplicitConsentRequired\020\001\022\020\n\014" +
      "OptoutCookie\020\002\022\023\n\017CtoOptoutCookie\020\003\022\025\n\021L" +
      "imitedAdTracking\020\004\022\016\n\nHstsOptout\020\005\022\024\n\020Do" +
      "NotTrackHeader\020\006\022\r\n\tOoOCookie\020\007\022\021\n\rPendi" +
      "ngOptout\020\010B\017\n\rcom.acme.glupb\006proto3"
    };
    descriptor = com.google.protobuf.Descriptors.FileDescriptor
      .internalBuildGeneratedFileFrom(descriptorData,
        new com.google.protobuf.Descriptors.FileDescriptor[] {
          com.acme.glup.MetadataProto.getDescriptor(),
        });
    internal_static_Acme_Glup_Consent_descriptor =
      getDescriptor().getMessageTypes().get(0);
    internal_static_Acme_Glup_Consent_fieldAccessorTable = new
      com.google.protobuf.GeneratedMessageV3.FieldAccessorTable(
        internal_static_Acme_Glup_Consent_descriptor,
        new java.lang.String[] { "IdentificationForbidden", "Reason", "SetFields", });
    internal_static_Acme_Glup_Consent_SetFieldsEntry_descriptor =
      internal_static_Acme_Glup_Consent_descriptor.getNestedTypes().get(0);
    internal_static_Acme_Glup_Consent_SetFieldsEntry_fieldAccessorTable = new
      com.google.protobuf.GeneratedMessageV3.FieldAccessorTable(
        internal_static_Acme_Glup_Consent_SetFieldsEntry_descriptor,
        new java.lang.String[] { "Key", "Value", });
    com.google.protobuf.ExtensionRegistry registry =
        com.google.protobuf.ExtensionRegistry.newInstance();
    registry.add(com.acme.glup.MetadataProto.containsNullableFields);
    com.google.protobuf.Descriptors.FileDescriptor
        .internalUpdateFileDescriptor(descriptor, registry);
    com.acme.glup.MetadataProto.getDescriptor();
  }

  // @@protoc_insertion_point(outer_class_scope)
}
