// Generated by the protocol buffer compiler.  DO NOT EDIT!
// source: TestOptionalProto.proto

package io.confluent.kafka.serializers.protobuf.test;

public final class TestMessageOptionalProtos {
  private TestMessageOptionalProtos() {}
  public static void registerAllExtensions(
      com.google.protobuf.ExtensionRegistryLite registry) {
  }

  public static void registerAllExtensions(
      com.google.protobuf.ExtensionRegistry registry) {
    registerAllExtensions(
        (com.google.protobuf.ExtensionRegistryLite) registry);
  }
  public interface TestMessageOptionalOrBuilder extends
      // @@protoc_insertion_point(interface_extends:io.confluent.kafka.serializers.protobuf.test.TestMessageOptional)
      com.google.protobuf.MessageOrBuilder {

    /**
     * <code>string test_string = 1;</code>
     * @return The testString.
     */
    java.lang.String getTestString();
    /**
     * <code>string test_string = 1;</code>
     * @return The bytes for testString.
     */
    com.google.protobuf.ByteString
        getTestStringBytes();

    /**
     * <code>optional string test_optional_string = 2;</code>
     * @return Whether the testOptionalString field is set.
     */
    boolean hasTestOptionalString();
    /**
     * <code>optional string test_optional_string = 2;</code>
     * @return The testOptionalString.
     */
    java.lang.String getTestOptionalString();
    /**
     * <code>optional string test_optional_string = 2;</code>
     * @return The bytes for testOptionalString.
     */
    com.google.protobuf.ByteString
        getTestOptionalStringBytes();
  }
  /**
   * Protobuf type {@code io.confluent.kafka.serializers.protobuf.test.TestMessageOptional}
   */
  public static final class TestMessageOptional extends
      com.google.protobuf.GeneratedMessageV3 implements
      // @@protoc_insertion_point(message_implements:io.confluent.kafka.serializers.protobuf.test.TestMessageOptional)
      TestMessageOptionalOrBuilder {
  private static final long serialVersionUID = 0L;
    // Use TestMessageOptional.newBuilder() to construct.
    private TestMessageOptional(com.google.protobuf.GeneratedMessageV3.Builder<?> builder) {
      super(builder);
    }
    private TestMessageOptional() {
      testString_ = "";
      testOptionalString_ = "";
    }

    @java.lang.Override
    @SuppressWarnings({"unused"})
    protected java.lang.Object newInstance(
        UnusedPrivateParameter unused) {
      return new TestMessageOptional();
    }

    @java.lang.Override
    public final com.google.protobuf.UnknownFieldSet
    getUnknownFields() {
      return this.unknownFields;
    }
    private TestMessageOptional(
        com.google.protobuf.CodedInputStream input,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws com.google.protobuf.InvalidProtocolBufferException {
      this();
      if (extensionRegistry == null) {
        throw new java.lang.NullPointerException();
      }
      int mutable_bitField0_ = 0;
      com.google.protobuf.UnknownFieldSet.Builder unknownFields =
          com.google.protobuf.UnknownFieldSet.newBuilder();
      try {
        boolean done = false;
        while (!done) {
          int tag = input.readTag();
          switch (tag) {
            case 0:
              done = true;
              break;
            case 10: {
              java.lang.String s = input.readStringRequireUtf8();

              testString_ = s;
              break;
            }
            case 18: {
              java.lang.String s = input.readStringRequireUtf8();
              bitField0_ |= 0x00000001;
              testOptionalString_ = s;
              break;
            }
            default: {
              if (!parseUnknownField(
                  input, unknownFields, extensionRegistry, tag)) {
                done = true;
              }
              break;
            }
          }
        }
      } catch (com.google.protobuf.InvalidProtocolBufferException e) {
        throw e.setUnfinishedMessage(this);
      } catch (java.io.IOException e) {
        throw new com.google.protobuf.InvalidProtocolBufferException(
            e).setUnfinishedMessage(this);
      } finally {
        this.unknownFields = unknownFields.build();
        makeExtensionsImmutable();
      }
    }
    public static final com.google.protobuf.Descriptors.Descriptor
        getDescriptor() {
      return io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.internal_static_io_confluent_kafka_serializers_protobuf_test_TestMessageOptional_descriptor;
    }

    @java.lang.Override
    protected com.google.protobuf.GeneratedMessageV3.FieldAccessorTable
        internalGetFieldAccessorTable() {
      return io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.internal_static_io_confluent_kafka_serializers_protobuf_test_TestMessageOptional_fieldAccessorTable
          .ensureFieldAccessorsInitialized(
              io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional.class, io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional.Builder.class);
    }

    private int bitField0_;
    public static final int TEST_STRING_FIELD_NUMBER = 1;
    private volatile java.lang.Object testString_;
    /**
     * <code>string test_string = 1;</code>
     * @return The testString.
     */
    @java.lang.Override
    public java.lang.String getTestString() {
      java.lang.Object ref = testString_;
      if (ref instanceof java.lang.String) {
        return (java.lang.String) ref;
      } else {
        com.google.protobuf.ByteString bs = 
            (com.google.protobuf.ByteString) ref;
        java.lang.String s = bs.toStringUtf8();
        testString_ = s;
        return s;
      }
    }
    /**
     * <code>string test_string = 1;</code>
     * @return The bytes for testString.
     */
    @java.lang.Override
    public com.google.protobuf.ByteString
        getTestStringBytes() {
      java.lang.Object ref = testString_;
      if (ref instanceof java.lang.String) {
        com.google.protobuf.ByteString b = 
            com.google.protobuf.ByteString.copyFromUtf8(
                (java.lang.String) ref);
        testString_ = b;
        return b;
      } else {
        return (com.google.protobuf.ByteString) ref;
      }
    }

    public static final int TEST_OPTIONAL_STRING_FIELD_NUMBER = 2;
    private volatile java.lang.Object testOptionalString_;
    /**
     * <code>optional string test_optional_string = 2;</code>
     * @return Whether the testOptionalString field is set.
     */
    @java.lang.Override
    public boolean hasTestOptionalString() {
      return ((bitField0_ & 0x00000001) != 0);
    }
    /**
     * <code>optional string test_optional_string = 2;</code>
     * @return The testOptionalString.
     */
    @java.lang.Override
    public java.lang.String getTestOptionalString() {
      java.lang.Object ref = testOptionalString_;
      if (ref instanceof java.lang.String) {
        return (java.lang.String) ref;
      } else {
        com.google.protobuf.ByteString bs = 
            (com.google.protobuf.ByteString) ref;
        java.lang.String s = bs.toStringUtf8();
        testOptionalString_ = s;
        return s;
      }
    }
    /**
     * <code>optional string test_optional_string = 2;</code>
     * @return The bytes for testOptionalString.
     */
    @java.lang.Override
    public com.google.protobuf.ByteString
        getTestOptionalStringBytes() {
      java.lang.Object ref = testOptionalString_;
      if (ref instanceof java.lang.String) {
        com.google.protobuf.ByteString b = 
            com.google.protobuf.ByteString.copyFromUtf8(
                (java.lang.String) ref);
        testOptionalString_ = b;
        return b;
      } else {
        return (com.google.protobuf.ByteString) ref;
      }
    }

    private byte memoizedIsInitialized = -1;
    @java.lang.Override
    public final boolean isInitialized() {
      byte isInitialized = memoizedIsInitialized;
      if (isInitialized == 1) return true;
      if (isInitialized == 0) return false;

      memoizedIsInitialized = 1;
      return true;
    }

    @java.lang.Override
    public void writeTo(com.google.protobuf.CodedOutputStream output)
                        throws java.io.IOException {
      if (!getTestStringBytes().isEmpty()) {
        com.google.protobuf.GeneratedMessageV3.writeString(output, 1, testString_);
      }
      if (((bitField0_ & 0x00000001) != 0)) {
        com.google.protobuf.GeneratedMessageV3.writeString(output, 2, testOptionalString_);
      }
      unknownFields.writeTo(output);
    }

    @java.lang.Override
    public int getSerializedSize() {
      int size = memoizedSize;
      if (size != -1) return size;

      size = 0;
      if (!getTestStringBytes().isEmpty()) {
        size += com.google.protobuf.GeneratedMessageV3.computeStringSize(1, testString_);
      }
      if (((bitField0_ & 0x00000001) != 0)) {
        size += com.google.protobuf.GeneratedMessageV3.computeStringSize(2, testOptionalString_);
      }
      size += unknownFields.getSerializedSize();
      memoizedSize = size;
      return size;
    }

    @java.lang.Override
    public boolean equals(final java.lang.Object obj) {
      if (obj == this) {
       return true;
      }
      if (!(obj instanceof io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional)) {
        return super.equals(obj);
      }
      io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional other = (io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional) obj;

      if (!getTestString()
          .equals(other.getTestString())) return false;
      if (hasTestOptionalString() != other.hasTestOptionalString()) return false;
      if (hasTestOptionalString()) {
        if (!getTestOptionalString()
            .equals(other.getTestOptionalString())) return false;
      }
      if (!unknownFields.equals(other.unknownFields)) return false;
      return true;
    }

    @java.lang.Override
    public int hashCode() {
      if (memoizedHashCode != 0) {
        return memoizedHashCode;
      }
      int hash = 41;
      hash = (19 * hash) + getDescriptor().hashCode();
      hash = (37 * hash) + TEST_STRING_FIELD_NUMBER;
      hash = (53 * hash) + getTestString().hashCode();
      if (hasTestOptionalString()) {
        hash = (37 * hash) + TEST_OPTIONAL_STRING_FIELD_NUMBER;
        hash = (53 * hash) + getTestOptionalString().hashCode();
      }
      hash = (29 * hash) + unknownFields.hashCode();
      memoizedHashCode = hash;
      return hash;
    }

    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parseFrom(
        java.nio.ByteBuffer data)
        throws com.google.protobuf.InvalidProtocolBufferException {
      return PARSER.parseFrom(data);
    }
    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parseFrom(
        java.nio.ByteBuffer data,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws com.google.protobuf.InvalidProtocolBufferException {
      return PARSER.parseFrom(data, extensionRegistry);
    }
    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parseFrom(
        com.google.protobuf.ByteString data)
        throws com.google.protobuf.InvalidProtocolBufferException {
      return PARSER.parseFrom(data);
    }
    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parseFrom(
        com.google.protobuf.ByteString data,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws com.google.protobuf.InvalidProtocolBufferException {
      return PARSER.parseFrom(data, extensionRegistry);
    }
    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parseFrom(byte[] data)
        throws com.google.protobuf.InvalidProtocolBufferException {
      return PARSER.parseFrom(data);
    }
    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parseFrom(
        byte[] data,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws com.google.protobuf.InvalidProtocolBufferException {
      return PARSER.parseFrom(data, extensionRegistry);
    }
    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parseFrom(java.io.InputStream input)
        throws java.io.IOException {
      return com.google.protobuf.GeneratedMessageV3
          .parseWithIOException(PARSER, input);
    }
    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parseFrom(
        java.io.InputStream input,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws java.io.IOException {
      return com.google.protobuf.GeneratedMessageV3
          .parseWithIOException(PARSER, input, extensionRegistry);
    }
    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parseDelimitedFrom(java.io.InputStream input)
        throws java.io.IOException {
      return com.google.protobuf.GeneratedMessageV3
          .parseDelimitedWithIOException(PARSER, input);
    }
    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parseDelimitedFrom(
        java.io.InputStream input,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws java.io.IOException {
      return com.google.protobuf.GeneratedMessageV3
          .parseDelimitedWithIOException(PARSER, input, extensionRegistry);
    }
    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parseFrom(
        com.google.protobuf.CodedInputStream input)
        throws java.io.IOException {
      return com.google.protobuf.GeneratedMessageV3
          .parseWithIOException(PARSER, input);
    }
    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parseFrom(
        com.google.protobuf.CodedInputStream input,
        com.google.protobuf.ExtensionRegistryLite extensionRegistry)
        throws java.io.IOException {
      return com.google.protobuf.GeneratedMessageV3
          .parseWithIOException(PARSER, input, extensionRegistry);
    }

    @java.lang.Override
    public Builder newBuilderForType() { return newBuilder(); }
    public static Builder newBuilder() {
      return DEFAULT_INSTANCE.toBuilder();
    }
    public static Builder newBuilder(io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional prototype) {
      return DEFAULT_INSTANCE.toBuilder().mergeFrom(prototype);
    }
    @java.lang.Override
    public Builder toBuilder() {
      return this == DEFAULT_INSTANCE
          ? new Builder() : new Builder().mergeFrom(this);
    }

    @java.lang.Override
    protected Builder newBuilderForType(
        com.google.protobuf.GeneratedMessageV3.BuilderParent parent) {
      Builder builder = new Builder(parent);
      return builder;
    }
    /**
     * Protobuf type {@code io.confluent.kafka.serializers.protobuf.test.TestMessageOptional}
     */
    public static final class Builder extends
        com.google.protobuf.GeneratedMessageV3.Builder<Builder> implements
        // @@protoc_insertion_point(builder_implements:io.confluent.kafka.serializers.protobuf.test.TestMessageOptional)
        io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptionalOrBuilder {
      public static final com.google.protobuf.Descriptors.Descriptor
          getDescriptor() {
        return io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.internal_static_io_confluent_kafka_serializers_protobuf_test_TestMessageOptional_descriptor;
      }

      @java.lang.Override
      protected com.google.protobuf.GeneratedMessageV3.FieldAccessorTable
          internalGetFieldAccessorTable() {
        return io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.internal_static_io_confluent_kafka_serializers_protobuf_test_TestMessageOptional_fieldAccessorTable
            .ensureFieldAccessorsInitialized(
                io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional.class, io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional.Builder.class);
      }

      // Construct using io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional.newBuilder()
      private Builder() {
        maybeForceBuilderInitialization();
      }

      private Builder(
          com.google.protobuf.GeneratedMessageV3.BuilderParent parent) {
        super(parent);
        maybeForceBuilderInitialization();
      }
      private void maybeForceBuilderInitialization() {
        if (com.google.protobuf.GeneratedMessageV3
                .alwaysUseFieldBuilders) {
        }
      }
      @java.lang.Override
      public Builder clear() {
        super.clear();
        testString_ = "";

        testOptionalString_ = "";
        bitField0_ = (bitField0_ & ~0x00000001);
        return this;
      }

      @java.lang.Override
      public com.google.protobuf.Descriptors.Descriptor
          getDescriptorForType() {
        return io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.internal_static_io_confluent_kafka_serializers_protobuf_test_TestMessageOptional_descriptor;
      }

      @java.lang.Override
      public io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional getDefaultInstanceForType() {
        return io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional.getDefaultInstance();
      }

      @java.lang.Override
      public io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional build() {
        io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional result = buildPartial();
        if (!result.isInitialized()) {
          throw newUninitializedMessageException(result);
        }
        return result;
      }

      @java.lang.Override
      public io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional buildPartial() {
        io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional result = new io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional(this);
        int from_bitField0_ = bitField0_;
        int to_bitField0_ = 0;
        result.testString_ = testString_;
        if (((from_bitField0_ & 0x00000001) != 0)) {
          to_bitField0_ |= 0x00000001;
        }
        result.testOptionalString_ = testOptionalString_;
        result.bitField0_ = to_bitField0_;
        onBuilt();
        return result;
      }

      @java.lang.Override
      public Builder clone() {
        return super.clone();
      }
      @java.lang.Override
      public Builder setField(
          com.google.protobuf.Descriptors.FieldDescriptor field,
          java.lang.Object value) {
        return super.setField(field, value);
      }
      @java.lang.Override
      public Builder clearField(
          com.google.protobuf.Descriptors.FieldDescriptor field) {
        return super.clearField(field);
      }
      @java.lang.Override
      public Builder clearOneof(
          com.google.protobuf.Descriptors.OneofDescriptor oneof) {
        return super.clearOneof(oneof);
      }
      @java.lang.Override
      public Builder setRepeatedField(
          com.google.protobuf.Descriptors.FieldDescriptor field,
          int index, java.lang.Object value) {
        return super.setRepeatedField(field, index, value);
      }
      @java.lang.Override
      public Builder addRepeatedField(
          com.google.protobuf.Descriptors.FieldDescriptor field,
          java.lang.Object value) {
        return super.addRepeatedField(field, value);
      }
      @java.lang.Override
      public Builder mergeFrom(com.google.protobuf.Message other) {
        if (other instanceof io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional) {
          return mergeFrom((io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional)other);
        } else {
          super.mergeFrom(other);
          return this;
        }
      }

      public Builder mergeFrom(io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional other) {
        if (other == io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional.getDefaultInstance()) return this;
        if (!other.getTestString().isEmpty()) {
          testString_ = other.testString_;
          onChanged();
        }
        if (other.hasTestOptionalString()) {
          bitField0_ |= 0x00000001;
          testOptionalString_ = other.testOptionalString_;
          onChanged();
        }
        this.mergeUnknownFields(other.unknownFields);
        onChanged();
        return this;
      }

      @java.lang.Override
      public final boolean isInitialized() {
        return true;
      }

      @java.lang.Override
      public Builder mergeFrom(
          com.google.protobuf.CodedInputStream input,
          com.google.protobuf.ExtensionRegistryLite extensionRegistry)
          throws java.io.IOException {
        io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional parsedMessage = null;
        try {
          parsedMessage = PARSER.parsePartialFrom(input, extensionRegistry);
        } catch (com.google.protobuf.InvalidProtocolBufferException e) {
          parsedMessage = (io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional) e.getUnfinishedMessage();
          throw e.unwrapIOException();
        } finally {
          if (parsedMessage != null) {
            mergeFrom(parsedMessage);
          }
        }
        return this;
      }
      private int bitField0_;

      private java.lang.Object testString_ = "";
      /**
       * <code>string test_string = 1;</code>
       * @return The testString.
       */
      public java.lang.String getTestString() {
        java.lang.Object ref = testString_;
        if (!(ref instanceof java.lang.String)) {
          com.google.protobuf.ByteString bs =
              (com.google.protobuf.ByteString) ref;
          java.lang.String s = bs.toStringUtf8();
          testString_ = s;
          return s;
        } else {
          return (java.lang.String) ref;
        }
      }
      /**
       * <code>string test_string = 1;</code>
       * @return The bytes for testString.
       */
      public com.google.protobuf.ByteString
          getTestStringBytes() {
        java.lang.Object ref = testString_;
        if (ref instanceof String) {
          com.google.protobuf.ByteString b = 
              com.google.protobuf.ByteString.copyFromUtf8(
                  (java.lang.String) ref);
          testString_ = b;
          return b;
        } else {
          return (com.google.protobuf.ByteString) ref;
        }
      }
      /**
       * <code>string test_string = 1;</code>
       * @param value The testString to set.
       * @return This builder for chaining.
       */
      public Builder setTestString(
          java.lang.String value) {
        if (value == null) {
    throw new NullPointerException();
  }
  
        testString_ = value;
        onChanged();
        return this;
      }
      /**
       * <code>string test_string = 1;</code>
       * @return This builder for chaining.
       */
      public Builder clearTestString() {
        
        testString_ = getDefaultInstance().getTestString();
        onChanged();
        return this;
      }
      /**
       * <code>string test_string = 1;</code>
       * @param value The bytes for testString to set.
       * @return This builder for chaining.
       */
      public Builder setTestStringBytes(
          com.google.protobuf.ByteString value) {
        if (value == null) {
    throw new NullPointerException();
  }
  checkByteStringIsUtf8(value);
        
        testString_ = value;
        onChanged();
        return this;
      }

      private java.lang.Object testOptionalString_ = "";
      /**
       * <code>optional string test_optional_string = 2;</code>
       * @return Whether the testOptionalString field is set.
       */
      public boolean hasTestOptionalString() {
        return ((bitField0_ & 0x00000001) != 0);
      }
      /**
       * <code>optional string test_optional_string = 2;</code>
       * @return The testOptionalString.
       */
      public java.lang.String getTestOptionalString() {
        java.lang.Object ref = testOptionalString_;
        if (!(ref instanceof java.lang.String)) {
          com.google.protobuf.ByteString bs =
              (com.google.protobuf.ByteString) ref;
          java.lang.String s = bs.toStringUtf8();
          testOptionalString_ = s;
          return s;
        } else {
          return (java.lang.String) ref;
        }
      }
      /**
       * <code>optional string test_optional_string = 2;</code>
       * @return The bytes for testOptionalString.
       */
      public com.google.protobuf.ByteString
          getTestOptionalStringBytes() {
        java.lang.Object ref = testOptionalString_;
        if (ref instanceof String) {
          com.google.protobuf.ByteString b = 
              com.google.protobuf.ByteString.copyFromUtf8(
                  (java.lang.String) ref);
          testOptionalString_ = b;
          return b;
        } else {
          return (com.google.protobuf.ByteString) ref;
        }
      }
      /**
       * <code>optional string test_optional_string = 2;</code>
       * @param value The testOptionalString to set.
       * @return This builder for chaining.
       */
      public Builder setTestOptionalString(
          java.lang.String value) {
        if (value == null) {
    throw new NullPointerException();
  }
  bitField0_ |= 0x00000001;
        testOptionalString_ = value;
        onChanged();
        return this;
      }
      /**
       * <code>optional string test_optional_string = 2;</code>
       * @return This builder for chaining.
       */
      public Builder clearTestOptionalString() {
        bitField0_ = (bitField0_ & ~0x00000001);
        testOptionalString_ = getDefaultInstance().getTestOptionalString();
        onChanged();
        return this;
      }
      /**
       * <code>optional string test_optional_string = 2;</code>
       * @param value The bytes for testOptionalString to set.
       * @return This builder for chaining.
       */
      public Builder setTestOptionalStringBytes(
          com.google.protobuf.ByteString value) {
        if (value == null) {
    throw new NullPointerException();
  }
  checkByteStringIsUtf8(value);
        bitField0_ |= 0x00000001;
        testOptionalString_ = value;
        onChanged();
        return this;
      }
      @java.lang.Override
      public final Builder setUnknownFields(
          final com.google.protobuf.UnknownFieldSet unknownFields) {
        return super.setUnknownFields(unknownFields);
      }

      @java.lang.Override
      public final Builder mergeUnknownFields(
          final com.google.protobuf.UnknownFieldSet unknownFields) {
        return super.mergeUnknownFields(unknownFields);
      }


      // @@protoc_insertion_point(builder_scope:io.confluent.kafka.serializers.protobuf.test.TestMessageOptional)
    }

    // @@protoc_insertion_point(class_scope:io.confluent.kafka.serializers.protobuf.test.TestMessageOptional)
    private static final io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional DEFAULT_INSTANCE;
    static {
      DEFAULT_INSTANCE = new io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional();
    }

    public static io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional getDefaultInstance() {
      return DEFAULT_INSTANCE;
    }

    private static final com.google.protobuf.Parser<TestMessageOptional>
        PARSER = new com.google.protobuf.AbstractParser<TestMessageOptional>() {
      @java.lang.Override
      public TestMessageOptional parsePartialFrom(
          com.google.protobuf.CodedInputStream input,
          com.google.protobuf.ExtensionRegistryLite extensionRegistry)
          throws com.google.protobuf.InvalidProtocolBufferException {
        return new TestMessageOptional(input, extensionRegistry);
      }
    };

    public static com.google.protobuf.Parser<TestMessageOptional> parser() {
      return PARSER;
    }

    @java.lang.Override
    public com.google.protobuf.Parser<TestMessageOptional> getParserForType() {
      return PARSER;
    }

    @java.lang.Override
    public io.confluent.kafka.serializers.protobuf.test.TestMessageOptionalProtos.TestMessageOptional getDefaultInstanceForType() {
      return DEFAULT_INSTANCE;
    }

  }

  private static final com.google.protobuf.Descriptors.Descriptor
    internal_static_io_confluent_kafka_serializers_protobuf_test_TestMessageOptional_descriptor;
  private static final 
    com.google.protobuf.GeneratedMessageV3.FieldAccessorTable
      internal_static_io_confluent_kafka_serializers_protobuf_test_TestMessageOptional_fieldAccessorTable;

  public static com.google.protobuf.Descriptors.FileDescriptor
      getDescriptor() {
    return descriptor;
  }
  private static  com.google.protobuf.Descriptors.FileDescriptor
      descriptor;
  static {
    java.lang.String[] descriptorData = {
      "\n\027TestOptionalProto.proto\022,io.confluent." +
      "kafka.serializers.protobuf.test\"f\n\023TestM" +
      "essageOptional\022\023\n\013test_string\030\001 \001(\t\022!\n\024t" +
      "est_optional_string\030\002 \001(\tH\000\210\001\001B\027\n\025_test_" +
      "optional_stringBI\n,io.confluent.kafka.se" +
      "rializers.protobuf.testB\031TestMessageOpti" +
      "onalProtosb\006proto3"
    };
    descriptor = com.google.protobuf.Descriptors.FileDescriptor
      .internalBuildGeneratedFileFrom(descriptorData,
        new com.google.protobuf.Descriptors.FileDescriptor[] {
        });
    internal_static_io_confluent_kafka_serializers_protobuf_test_TestMessageOptional_descriptor =
      getDescriptor().getMessageTypes().get(0);
    internal_static_io_confluent_kafka_serializers_protobuf_test_TestMessageOptional_fieldAccessorTable = new
      com.google.protobuf.GeneratedMessageV3.FieldAccessorTable(
        internal_static_io_confluent_kafka_serializers_protobuf_test_TestMessageOptional_descriptor,
        new java.lang.String[] { "TestString", "TestOptionalString", "TestOptionalString", });
  }

  // @@protoc_insertion_point(outer_class_scope)
}
