syntax = "proto3";

import "confluent/meta.proto";
import "confluent/type/decimal.proto";

option java_package = "io.confluent.kafka.serializers.protobuf.test";

// Wrapper message for `Decimal`.
message DecimalValue {
  option (confluent.message_meta).doc = "message";

  // The bytes value.
  confluent.type.Decimal value = 1 [(confluent.field_meta) = { params: [
    { key: "precision", value: "8" },
    { key: "scale", value: "3" }
  ]}];
}
