syntax = "proto3";

import "google/type/date.proto";

option java_package = "io.confluent.kafka.serializers.protobuf.test";

// Wrapper message for `Date`.
message DateValue {
  // The bytes value.
  google.type.Date value = 1;
}
